/**********************************************************************/
//                                                                    //
// /****************************************************************/ //
// /*                                                              */ //
// /*                         AMBIPAN                              */ //
// /*                                                              */ //
// /* Auteur: Rmi MIGNOT                                          */ //
// /*         Elve ingnieur Tlcom2  l'ISPG                    */ //
// /*         (Institut Suprieur Polytechnique de Galile),       */ //
// /*         Universit Paris13.                                  */ //
// /*                                                              */ //
// /* Date de cration:   11/07/03                                 */ //
// /* Version: 1.5        17/07/04                                 */ //
// /*                                                              */ //
// /* Ralis  la MSH Paris Nord (Maison des Sciences de l'Homme) */ //
// /*         en collaboration avec A.Sedes, B.Courribet           */ //
// /*         et J.B.Thiebaut,                                     */ //
// /*         CICM Universit Paris8, MSH Paris Nord,              */ //
// /*         ACI Jeunes Chercheurs "Espaces Sonores".             */ //
// /*                                                              */ //
// /****************************************************************/ //
//                                                                    //
/**********************************************************************/


/**
 * Copyright (C) 2003-2004 Rmi Mignot, MSH Paris Nord,
 * 
 * This library is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU Library General Public License as published 
 * by the Free Software Foundation; either version 2 of the License.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public 
 * License for more details.
 * 
 * You should have received a copy of the GNU Library General Public License 
 * along with this library; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * www.mshparisnord.org
 * rmignot@mshparisnord.org
 */


/*Description:  Cet objet permet de spatialiser un son monoral
                 l'aide de N haut-parleurs situs en cercle autour
                de l'auditeur. La spatialisation se fait grce 
                l'ambisonie de Michael Gerzon.                        */ 



/**********************************************************************/
/*                            EN TETE                                 */
/**********************************************************************/
 
#include <math.h>
#include "ext.h"
#include "z_dsp.h"

#define   Nmax      16     //Nombre maximum de haut-parleurs,
#define   Ndefaut   4      //Nombre de haut-parleurs par dfaut,
#define   Xdefaut   0      //Valeur par defaut de l'absisse,
#define   Ydefaut   1      //Valeur par dfaut de l'ordonne,
#define   Phidefaut 1.5708F//Valeur par dfaut de l'angle (PI/2),
#define   Rdefaut   1      //Distance par dfaut de la source,
#define   OFFSET    .3     //Offset par defaut de l'ambisonie,
#define   T_COS     4096   //Taille du tableau cosinus (puissance de 2),
#define   MASQUE    4095   //Masque pour le modulo = T_COS - 1,
#define   DTIME     10     //Temps d'interpolation par dfaut,
                           //en chantillons (10ms),
#define   Pi        3.1415926535897932384F  // Pi
#define   I360      0.0027777777777777778F  // 1/360
#define   EPSILON   0.000001 //Critre pour tester si les flottants sont nuls



/**********************************************************/
/*         La classe                                      */
/**********************************************************/

void *ambipan_tilde_class;

typedef struct _ambipan_tilde
{  
  t_pxobject x_obj;     //Membre indispensable  Pure Data,
  
  int     base;         //Type de base, 1->cartsienne, 0->polaire, 		
  int     controle;     //1->coordonnes en contrle, 0->coordonnes en signal,
  int     mute;         //1->entres signal mutes, 0->non mutes,
  int     Nout;         //Nombre de sorties,
  int     N;            //Nombre de haut-parleurs.
 
  float   x, y;         //Coordonnes de la source en cartsien,
  float   phi, r;       //Coordonnes polaires de la source,

  float   offset;       //Offset de l'ambisonie,
  int     dtime;        //Temps d'interpolation en chantillons,
 
  float   P[Nmax];      //Tableau contenant les N coefficients ambisoniques Pn,
  float   teta[Nmax];   //Angle de chaque haut-parleur,
  float   dist[Nmax];   //Distance des haut-parleurs,

  float   dP[Nmax];     //pas pour l'interpolation,
  float   Pstop[Nmax];  //cible pour l'interpolation,

  float   *cosin;       //Adresse des tableaux pour les cosinus,
  float   *cos_teta;    //les sinus et cosinus des angles des 
  float   *sin_teta;    //haut-parleurs.
 
} t_ambipan_tilde;



  
/**********************************************************************/
/*      Prototypage des fonctions et mthodes.     	                  */
/**********************************************************************/

void  ambipan_tilde_dsp(t_ambipan_tilde *x, t_signal **sp, short *count);
t_int *ambipan_tilde_perform_signal(t_int *w);
t_int *ambipan_tilde_perform_controle(t_int *w);
void  ambipan_tilde_recoit_x(t_ambipan_tilde *x, double xp);
void  ambipan_tilde_recoit_y(t_ambipan_tilde *x, double yp);
void  ambipan_tilde_recoit_liste( t_ambipan_tilde *x, t_symbol *s, int argc, t_atom *argv);
void  ambipan_tilde_initialiser_nb_hp( t_ambipan_tilde *x, long N);
void  ambipan_tilde_dist_teta_positionner_hp( t_ambipan_tilde *x, t_symbol *s, int argc, t_atom *argv );
void  ambipan_tilde_teta_positionner_hp( t_ambipan_tilde *x, t_symbol *s, int argc, t_atom *argv );
void  ambipan_tilde_xy_positionner_hp( t_ambipan_tilde *x, t_symbol *s, int argc, t_atom *argv );
void  ambipan_tilde_changer_offset( t_ambipan_tilde *x, t_symbol *s, int argc, t_atom *argv );
void  ambipan_tilde_changer_type_repere(t_ambipan_tilde *x, t_symbol *sym);
void  ambipan_tilde_muter_entrees_signal(t_ambipan_tilde *x, int mute);
void  ambipan_tilde_informations( t_ambipan_tilde *x );
void  ambipan_tilde_dest(t_ambipan_tilde *x);
void  *ambipan_tilde_new( Symbol *s, int argc, Atom *argv );
void  main(void);




/**********************************************************************/
/*      TRAITEMENT DU SIGNAL avec coordonnes recues en controle.     */
/**********************************************************************/

t_int *ambipan_tilde_perform_controle(t_int *w)
{
  /*Prparation des pointeurs****************************************/
  t_ambipan_tilde *x = (t_ambipan_tilde *)(w[1]);//adresse de la dataspace
  float     *son_in   = (t_float *)(w[2]);       //son en entre     
  long      n         = (long)(w[3+x->Nout]);    //taille des blocs

  /*Dclarations des variables locales*/
  float son;                    //chantillon temporaire,
  int   hp;                     //hp = n du haut-parleur, 
  int   i;                      //i = indice des chantillons,
  int   N = x->N;               //Nombre de haut-parleur,
  int   retour = 4 + x->Nout;   //pour le pointeur de retour.

  /*Prparation des pointeurs de sorties ****************************/
  float *out[Nmax];
  for( hp = 0; hp < x->Nout; hp++)
    out[hp] = (t_float *)(w[3+hp]);



  /******************************************************************/
  /*  Traitements  **************************************************/

  for( i=0; i<n; i++)
  {
    //on stocke l'chantillon d'entre dans une variable temporaire
    //car son_in = out[0].
    son = son_in[i];


    //Modulation des sorties avec les coefficients ambisoniques Pn.
    for( hp = N-1; hp >= 0; hp--)
    {
      //Incrmentation des P pour l'interpolation,
      if( x->P[hp] == x->Pstop[hp] ) /*rien*/;
      else if ( fabs(x->Pstop[hp] - x->P[hp]) > fabs(x->dP[hp]) )
        x->P[hp] += x->dP[hp];
      else
        x->P[hp] = x->Pstop[hp];

      /******************************/
      out[hp][i] = son * x->P[hp];/**/
      /******************************/
    }
  }


  /*Initialisation  zro des sorties inutilises*/
  if( x->Nout > x->N){
    for( hp = x->Nout-1 ; hp >= N ; hp--){
      for( i=n-1 ; i>=0; i--)
        out[hp][i] = 0;
    }
  }
  
  return ( w+retour );
}



/**********************************************************************/
/*      TRAITEMENT DU SIGNAL avec coordonnes recues en signal.       */
/**********************************************************************/

t_int *ambipan_tilde_perform_signal(t_int *w)
{
  /*Prparation des pointeurs****************************************/
  t_ambipan_tilde *x = (t_ambipan_tilde *)(w[1]);//adresse de la dataspace
  float     *son_in   = (t_float *)(w[2]);       //son en entre
  float     *xp       = (t_float *)(w[3]);       //rception de x
  float     *yp       = (t_float *)(w[4]);       //rception de y
  int       n         = (int)(w[5+x->Nout]);     //taille des blocs

  /*Dclarations des variables locales*/
  float son;            //variable temporaire,
  int   hp;             //indices relatif au haut-parleur,
  int   i;              //indices du n de l'chantillon,
  int   N = x->N;       //Nombre de haut-parleur,_
  int   retour = x->Nout + 6; //retour pour pure data,
  float offset = x->offset;

  float K = (float)(sqrt(1/(float)N)/1.66);   //Facteur correctif.

  //Paramtres ambisoniques:
  float xtemp, xl, yl, ds, dist, X, Y, W, P;
  int   phii;

  /*Prparation des pointeurs des sorties ***************************/
  float *out[Nmax];
  for( hp = 0; hp < x->Nout; hp++)
    out[hp] = (t_sample *)(w[hp+5]);



  /******************************************************************/
  /*  Traitements  **************************************************/

  /******************************************************************/
  /*Si entrees signal mutes, on utilise le tableau P de contrle.***/
  if( x->mute ) 
    for( i=0; i<n; i++)
    {
      //on stocke l'chantillon d'entre dans une variable temporaire
      //car son_in = out[0].
      son = son_in[i];


      //Modulation des sorties avec les coefficients ambisoniques Pn.
      for( hp = N-1; hp >= 0; hp--)
      {
        //Incrmentation des P pour l'interpolation,
        if( x->P[hp] == x->Pstop[hp] ) /*rien*/  ;
        else if ( fabs(x->Pstop[hp] - x->P[hp]) > fabs(x->dP[hp]) )
          x->P[hp] += x->dP[hp];
        else
          x->P[hp] = x->Pstop[hp];

        /******************************/
        out[hp][i] = son * x->P[hp];/**/
        /******************************/
      }
    }
  
  /******************************************************************/
  /*Si entres signal non mutes, on utilise les vecteurs de signal**/
  else 
    for( i=n-1; i>=0; i--)
    {
      son = son_in[i];    //On stocke les chantillons d'entre, dans des
      xl = xp[i];         //constantes, car elles vont tre crases.
      yl = yp[i];
  
      //Conversion polaires -> cartsiennes,
      if( !x->base )
      {
        //ici xl = rayon et yl = angle, 
        phii = (int)( yl*T_COS*I360 )&(int)MASQUE; 
        xtemp    = xl*x->cosin[ phii ];
        phii = (int)(.25*T_COS - phii)&(int)MASQUE;
        yl       = xl*x->cosin[ phii ];
        xl = xtemp;
        //maintenant xl = abscisse et yl = ordonne.  
      }
  
      //Calcul des distances,
      ds   = xl*xl + yl*yl;
      dist = (float)sqrt(ds);

      //Calcul des paramtres ambisoniques,
      X = (float)( 2*xl / (ds + offset) ); 
      Y = (float)( 2*yl / (ds + offset) ); 
      W = (float)( .707 / (dist + offset) );

      for( hp=N-1; hp >= 0 ; hp--)
      {
        P = K  * ( W + X*x->cos_teta[hp]  
                     + Y*x->sin_teta[hp]  )
               * x->dist[hp];
      
        //Si Pn<0 on les force  0
        if(P < 0)            P = 0;

        /***********************/
        out[hp][i] = son * P;/**/
        /***********************/
      }
    }


  /*Initialisation  zro des sorties inutilises*/
  if( x->Nout > x->N){
    for( hp = x->Nout-1 ; hp >= N ; hp--){
      for( i=n-1 ; i>=0; i--)
        out[hp][i] = 0;
    }
  }

  return ( w+retour );
}



/**********************************************************************/
/*        METHODE RECOIT_x                                            */
/**********************************************************************/
//Cette mthode reoit x et change tous les paramtres ambisoniques: 
//  X, Y, W et les Pn 
void ambipan_tilde_recoit_x(t_ambipan_tilde *x, double xp)
{
  int hp;
  float ds, dist;
  float X, Y, W;
  float yp = x->y;
  float K = (float)(sqrt(1/(float)x->N)/1.66 ); 


  //Conversion polaire -> cartsienne,
  if( !x->base )
  {
    //ici xp = rayon,
    x->r = xp;
    xp   = (float)( x->r * cos( x->phi ) );
    yp   = (float)( x->r * sin( x->phi ) );
    //maintenant xp = abscisse,
  }
  x->x = xp;
  x->y = yp;

  //Calcul des distances,
  ds = xp*xp + yp*yp;
  dist = (float)sqrt(ds);

  //Calcul des paramtres ambisoniques,
  X = (float)( 2*xp/(ds + x->offset) ); 
  Y = (float)( 2*yp/(ds + x->offset) ); 
  W = (float)( .707/(dist + x->offset) );

  //Calcul des coefficients ambisoniques cibles et des pas,
  for( hp = x->N-1; hp >= 0 ; hp--)
  {
    x->Pstop[hp] = (float)( ( W + X*cos(x->teta[hp]) 
                                + Y*sin(x->teta[hp])
                            ) * K  * x->dist[hp]);
    //Si Pstop_n < 0 on les force  0.
    if(x->Pstop[hp] < 0)
      x->Pstop[hp] = 0;

    x->dP[hp] = (x->Pstop[hp] - x->P[hp])/(float)x->dtime;
  }

  return;
}



/**********************************************************************/
/*        METHODE RECOIT_y                                            */
/**********************************************************************/
//Cette mthode reoit y et change tous les paramtres ambisoniques: 
//  X, Y, W et les Pn 
void ambipan_tilde_recoit_y(t_ambipan_tilde *x, double yp)
{
  int hp;
  float ds, dist;
  float X, Y, W;
  float xp = x->x;
  float K = (float)(sqrt(1/(float)x->N)/1.66 ); 

  
  //Conversion polaires -> cartsienne,
  if( !x->base )
  {
    //ici yp = angle,
    x->phi = (float)( yp*Pi/180 );
    xp   = (float)( x->r * cos( x->phi ) );
    yp   = (float)( x->r * sin( x->phi ) );
    //maintenant xp = ordonne,
  }
  x->y = yp;
  x->x = xp;

  //Calcul des distances,
  ds = xp*xp + yp*yp;
  dist = (float)sqrt(ds);

  //Calcul des paramtres ambisoniques,
  X = (float)( 2*xp/(ds + x->offset) ); 
  Y = (float)( 2*yp/(ds + x->offset) ); 
  W = (float)( .707/(dist + x->offset) );

  //Calcul des coefficients ambisoniques cibles et des pas,
  for( hp=x->N-1; hp >= 0 ; hp--)
  {
    x->Pstop[hp] = (float)( ( W + X*cos(x->teta[hp]) 
                                + Y*sin(x->teta[hp])
                            ) * K  * x->dist[hp]);
    //Si Pstop_n<0 on les force  0
    if(x->Pstop[hp] < 0)
      x->Pstop[hp] = 0;

    x->dP[hp] = (x->Pstop[hp] - x->P[hp])/(float)x->dtime;    
  }

  return;
}



/**********************************************************************/
/*        METHODE RECOIT_LISTE                                        */
/**********************************************************************/

void ambipan_tilde_recoit_liste( t_ambipan_tilde *x, t_symbol *s, int argc, t_atom *argv)
{
  float xp, yp = 0;
  
  //Rcupration du premier paramtre (abscisse ou rayon),
  if( argv[0].a_type == A_FLOAT )
    xp = argv[0].a_w.w_float;
  else if( argv[0].a_type == A_LONG )
    xp = (float)(argv[0].a_w.w_long);
    
  //Rcupration du second paramtre (ordonne ou angle),
  if( argc >= 2 )
  {
    if( argv[1].a_type == A_FLOAT )
      yp = argv[1].a_w.w_float;
    else if( argv[1].a_type == A_LONG )
      yp = (float)(argv[1].a_w.w_long);
  } 
  //si qu'un paramtre, on prend la valeur dja prsente.
  else if( x->base )     yp = x->y;
  else                   yp = x->phi*180/Pi;
  
  //Initialisation des valeurs,
  if( x->base )    x->y = yp;
  else             x->phi = yp*Pi/180;

  ambipan_tilde_recoit_x( x, xp);

  return;
}



/**********************************************************************/
/*      METHODE POSITIONNER HAUT-PARLEUR sur le cercle avec phi       */
/**********************************************************************/

void ambipan_tilde_teta_positionner_hp( t_ambipan_tilde *x, t_symbol *s, int argc, t_atom *argv )
{
  int hp;

  for( hp=0; hp<x->N && hp<argc; hp++)
  {
    //Rcupration des angles.
    if( argv[hp].a_type == A_FLOAT )
      x->teta[hp] = ((argv[hp].a_w.w_float)*Pi/180.);
    else if( argv[hp].a_type == A_LONG )
      x->teta[hp] = (float)((argv[hp].a_w.w_long)*Pi/180.);
    else if( argv[hp].a_type == A_SYM ){
      post("ambipan~: erreur, les membres de la liste doivent tre des nombres.");
      return;
    }
    
    x->dist[hp] = 1;
   
    //Prcalcul des cos et sin en signal. 
    if( !x->controle )
    {
      x->cos_teta[hp] = (float)cos( x->teta[hp] );
      x->sin_teta[hp] = (float)sin( x->teta[hp] );
    }
  }
  
  /*Affectation des gains de l'ambisonie*/
  if(x->controle)
    if(x->base)     ambipan_tilde_recoit_x( x, x->x);
    else            ambipan_tilde_recoit_x( x, x->r);

  return;
}



/**********************************************************************/
/*      METHODE POSITIONNER HAUT-PARLEUR  avec phi et le rayon        */
/**********************************************************************/

void ambipan_tilde_dist_teta_positionner_hp( t_ambipan_tilde *x, t_symbol *s, int argc, t_atom *argv )
{
  int hp;

  for( hp=0; hp<x->N && 2*hp<argc; hp++)
  {
    //Rcupration des distances des haut-praleurs au centre,
    if( argv[2*hp].a_type == A_FLOAT )
      x->dist[hp] = (argv[2*hp].a_w.w_float);
    else if( argv[2*hp].a_type == A_LONG )
      x->dist[hp] = (float)(argv[2*hp].a_w.w_long);
    else if( argv[2*hp].a_type == A_SYM ){
      post("ambipan~: erreur, les membres de la liste doivent tre des nombres.");
      return;
    }

    //Rcupration des angles des hp,
    if( 2*hp+1<argc )
    {
      if( argv[2*hp+1].a_type == A_FLOAT )
        x->teta[hp] = ((argv[2*hp+1].a_w.w_float)*Pi/180);
      else if( argv[2*hp+1].a_type == A_LONG )
        x->teta[hp] = (float)((argv[2*hp+1].a_w.w_long)*Pi/180);
      else if( argv[2*hp+1].a_type == A_SYM ){
        post("ambipan~: erreur, les membres de la liste doivent tre des nombres.");
        return;
      }
    }
   
    //Prcalcul des cos et sin en signal.
    if( !x->controle )
    {
      x->cos_teta[hp] = (float)cos( x->teta[hp] );
      x->sin_teta[hp] = (float)sin( x->teta[hp] );
    }
  }
    
  /*Affectation des gains de l'ambisonie*/
  if(x->controle)
    if(x->base)     ambipan_tilde_recoit_x( x, x->x);
    else            ambipan_tilde_recoit_x( x, x->r);

  return;
}



/**********************************************************************/
/*      METHODE POSITIONNER HAUT-PARLEUR  avec coordonnes x et y     */
/**********************************************************************/

void ambipan_tilde_xy_positionner_hp( t_ambipan_tilde *x, t_symbol *s, int argc, t_atom *argv )
{
  int hp;
  float xp, yp = 0,
        rayon, angle;

  /*Rcupration des coordonnes cartsiennes**********/
  //Abscisses:
  for( hp=0; hp<x->N && 2*hp<argc; hp++)
  {
    if( argv[2*hp].a_type == A_FLOAT )
      xp = (argv[2*hp].a_w.w_float);
    else if( argv[2*hp].a_type == A_LONG )
      xp = (float)(argv[2*hp].a_w.w_long);
    else if( argv[2*hp].a_type == A_SYM ){
      post("ambipan~: erreur, les membres de la liste doivent tre des nombres.");
      return;
    }

    //Ordonnes:
    if( 2*hp+1<argc )
    {
      if( argv[2*hp+1].a_type == A_FLOAT )
        yp = (argv[2*hp+1].a_w.w_float);
      else if( argv[2*hp+1].a_type == A_LONG )
        yp = (float)(argv[2*hp+1].a_w.w_long);
      else if( argv[2*hp+1].a_type == A_SYM ){
        post("ambipan~: erreur, les membres de la liste doivent tre des nombres.");
        return;
      }
    }


    /*Conversion en polaires***************************/  
    rayon = (float)sqrt( pow( xp, 2) + pow( yp, 2) );
    if( rayon!=0 )
    {
      angle = (float)acos( (float)xp/rayon );
      if(yp<0)
        angle = -angle;
    }
    else
      angle = 0;

    //Affectatoin des nouvelles valeurs
    x->teta[hp] = angle;
    x->dist[hp] = rayon;  
   
    /*Prcalcul des cos et des sin ********************/
    if( !x->controle )
    {
      x->cos_teta[hp] = (float)(cos( angle ));
      x->sin_teta[hp] = (float)(sin( angle ));
    }
  }
    
  /*Affectation des gains de l'ambisonie*/
  if(x->controle)
    if(x->base)     ambipan_tilde_recoit_x( x, x->x);
    else            ambipan_tilde_recoit_x( x, x->r);
    
  return;
}



/**********************************************************************/
/*              METHODE MODIFIER LE NOMBRE DE HP                      */
/**********************************************************************/

void ambipan_tilde_initialiser_nb_hp( t_ambipan_tilde *x, long N)
{
  int hp;


  if( x->Nout >= N && 2 <= N )
  {
    x->N = (int)N;
    
    //Modifications des angles des haut-parleurs et des rayons,
    for( hp=0; hp<x->N; hp++)
    {
      x->teta[hp] = (float)( Pi*( .5 + (1 - 2*hp )/(float)x->N) );
      x->dist[hp] = 1;
    }


    /*****************************************************/
    /*Modification des tableaux cos_teta et sin_teta     */
    /*pour l'audio.                                      */
    if( !x->controle )
    {
      //Prcalculs des cos et sin des haut-parleurs.
      for( hp=0; hp<x->N; hp++)
      {
        x->cos_teta[hp] = (float)cos( x->teta[hp] );
        x->sin_teta[hp] = (float)sin( x->teta[hp] );
      }
    }
  }  
  else
    post("ambipan~: Le nombre de haut-parleurs doit tre "
                    "infrieur  %d, \n"
                    "  et suprieur  2, ou gal.", x->Nout);
    
  /*Affectation des gains de l'ambisonie*/
  if(x->controle)
    if(x->base)     ambipan_tilde_recoit_x( x, x->x);
    else            ambipan_tilde_recoit_x( x, x->r);
   
}



/**********************************************************************/
/*              METHODE POUR MUTER LES ENTREES SIGNAL                 */
/**********************************************************************/

void  ambipan_tilde_muter_entrees_signal(t_ambipan_tilde *x, int mute)
{
  if( mute == 0)   x->mute = 0;
  else             x->mute = 1;
}



/**********************************************************************/
/*                   MODIFICATION DE L'OFFSET                         */
/**********************************************************************/

void ambipan_tilde_changer_offset( t_ambipan_tilde *x, t_symbol *s, int argc, t_atom *argv)
{
  float offset;

  if( argv[0].a_type == A_FLOAT )
    offset = (argv[0].a_w.w_float);
  else if( argv[0].a_type == A_LONG )
    offset = (float)(argv[0].a_w.w_long);
  else if( argv[0].a_type == A_SYM ){
    post("ambipan~: erreur, l'offset doit tre un nombre.");
    return;
  }
    
  if( offset <= EPSILON ){
    post("ambipan~: Pas d'offset ngatif ou nul s'il vous plait.");
    return;
  }

  //Changement de l'offset
  x->offset = offset; 
  
  /*Affectation des gains de l'ambisonie*/
  if(x->controle)
    if(x->base)     ambipan_tilde_recoit_x( x, x->x);
    else            ambipan_tilde_recoit_x( x, x->r);
   
}



/**********************************************************************/
/*                   MODIFICATION DU TYPE DE REPERE                   */
/**********************************************************************/

void ambipan_tilde_changer_type_repere( t_ambipan_tilde *x, t_symbol *sym)
{
  //Changement de x->base
  if( sym->s_name[0] == 'c')
    x->base = 1;
  else if( sym->s_name[0] == 'p')
    x->base = 0; 
  else 
    post("ambipan~: erreur, ne reconnait le type de repre.");
    
  //Si coordonnes polaires il faut initialiser x->r et x->phi:
  if( x->base == 0 )
  {
    //Conversion de x et y en polaires:  
    x->r = (float)sqrt( pow( x->x, 2) + pow( x->y, 2) );
    if( x->r!=0 )
    {
      x->phi = (float)acos( (float)x->x/x->r );
      if(x->y<0)
        x->phi = -x->phi;
    }
    else
      x->phi = 0;
  }  
}



/**********************************************************************/
/*         AFFICHAGE DES INFORMATIONS DANS LA FENETRE DE MAX          */
/**********************************************************************/

void ambipan_tilde_informations( t_ambipan_tilde *x)
{
  int hp;
  
  post("Info Ambipan~: ");
  
  if(x->controle && x->base)
    post("   coordonnes cartsiennes en contrle,");
  else if(x->controle && !x->base)
    post("   coordonnes polaires en contrle,");
  else if(!x->controle && x->base)
    post("   coordonnes cartsiennes en signal,");
  else if(!x->controle && !x->base)
    post("   coordonnes polaires en signal,");
    
  if(!x->controle && x->mute )
    post("   avec entres signal inactives,");
  
  post("   offset                 = %f,", x->offset);
  if(x->controle)
    post("   temps d'interpolation  = %d ms,", 
         (int)(x->dtime*1000/sys_getsr()));
  post("   nombre d'haut-parleurs = %d," , x->N);
  
  post("   position des haut-parleurs:");
  for( hp=0; hp<x->N-1; hp++)
    post("      hp n%d: %f.x + %f.y,", hp+1, x->dist[hp]*cos(x->teta[hp]), 
                                            x->dist[hp]*sin(x->teta[hp]));
    
  post("      hp n%d: %f.x + %f.y.", hp+1, x->dist[hp]*cos(x->teta[hp]), 
                                          x->dist[hp]*sin(x->teta[hp]));
}
    


/**********************************************************************/
/*                       FONCTION DESTRUCTION                         */
/**********************************************************************/

void ambipan_tilde_dest(t_ambipan_tilde *x)
{
  //Libration de la mmoire alloue pour les cosinus,
  if( !x->controle  )
    freebytes( (void*)(x->cos_teta), (short)(2*x->N*sizeof(float)) );
 
  return;
}



/**********************************************************************/
/*                       FONCTION CREATION                            */
/**********************************************************************/

void *ambipan_tilde_new( t_symbol *s, int argc, t_atom *argv )
{  
  int    i, hp;
  char   car;

 
  /*Allocation de la dataspace******************************/
  t_ambipan_tilde *x = (t_ambipan_tilde *)newobject(ambipan_tilde_class);



  /*********************************************************/
  /*Rcupration du nombre de sorties et de haut-parleurs. */
  if( argc >= 1 ){
    if( argv[0].a_type == A_LONG )
      x->N = argv[0].a_w.w_long;
    else x->N = 0;               }
  else 
    x->N = Ndefaut;

  if( 16 < x->N || 2 > x->N ){
    x->N = Ndefaut;
    post("ambipan~: Il y a un problme dans la dclaration "
         "du nombre de haut-parleur, ");
    post("  il est de %d par dfaut.", Ndefaut);
  }
  x->Nout = x->N;  //Mme nombre de sorties et d'haut-parleur.
  
  /*Rcupration du type repre ************************/
  if( argc >=2 ){
    if( argv[1].a_type == A_SYM )
      car = (char)(argv[1].a_w.w_sym->s_name[0]);
    
    if( car == 'c' )
      x->base=1;
    else if( car == 'p' )
      x->base=0;
    else {
      x->base=1;
      post("ambipan~: erreur dans le type des coordonnes, ");
      post("  elles sont cartsiennes par dfaut.");
    }
  }
  else 
    x->base = 1; //Cartsienne par dfaut.
    
  /*Rcupration du type des entres *******************/
  if( argc >=3 ){
    if( argv[2].a_type == A_SYM )
      car = (char)(argv[2].a_w.w_sym->s_name[0]);

    if( car == 'c' )
      x->controle=1;
    else if( car == 's' )
      x->controle=0;
    else {
      x->controle=1;
      post("ambipan~: erreur dans le type des entres, ");
      post("  elles sont en controle par dfaut.");
    }
  }
  else 
    x->controle = 1;

  /*Rcupration de l'offset ***************************/
  if( argc >= 4 && argv[3].a_type == A_FLOAT )
    x->offset = fabs(argv[3].a_w.w_float);
  else if( argc >= 4 && argv[3].a_type == A_LONG )
    x->offset = fabs(argv[3].a_w.w_long);
  else
    x->offset = (float)OFFSET;
  if( x->offset <= EPSILON ){
    post("ambipan~: Pas d'offset ngatif ou nul s'il vous plait.");
    x->offset = (float)OFFSET;;
  }

  
  /*Rcupration du temps d'interpolation **************/
  if( argc >= 5 && argv[4].a_type == A_LONG )
    x->dtime = (int)(argv[4].a_w.w_long*sys_getsr()/1000.);
  else
    x->dtime = DTIME*sys_getsr()/1000.;
  if( x->dtime <= 0 ) //Pas de temps d'interpolation nul ou ngatif.
    x->dtime = 1;



  /*********************************************************/
  /*Initialisation des donnes de la classe. ***************/
  for( hp = x->N-1; hp >= 0; hp--) //Initialisation des P
    x->P[hp] = 0;

  x->mute= 0;             //entres signal non mutes
  x->y   = Ydefaut;       //initialisation de y
  x->phi = Phidefaut;     //initialisation de phi
         
  //Initialisation des angles des haut-parleurs et des rayons,
  for( hp=0; hp<x->N; hp++)
  {
    x->teta[hp] = (float)( Pi*( .5 + (1 - 2*hp )/(float)x->N) );
    x->dist[hp] = 1;
  }

  /*********************************************************/
  /* Cration des tableaux cosinus, cos_teta et sin_teta.  */
  /* pour l'audio.                                         */
  if( !x->controle )
  {
    x->cos_teta = (float*)getbytes( (short)((T_COS+2*x->N)*sizeof(float)) );
    x->sin_teta = x->cos_teta+x->N;

    //Prcalculs des cos et sin des haut-parleurs.
    for( hp=0; hp<x->N; hp++)
    {
      x->cos_teta[hp] = (float)cos( x->teta[hp] );
      x->sin_teta[hp] = (float)sin( x->teta[hp] );
    }

    //Remplissage du tableau cosinus,
    x->cosin = x->sin_teta+x->N;
    /*Pour avoir besoin de cos( phi ), on cherche:
      cos(phi) = 
      cosin[ (int)( phi*T_COS/360 ))&(((int)T_COS-1) ] */
    for( i=0; i<T_COS; i++)
      x->cosin[i] = (float)cos( i*2*Pi/T_COS );
  }

  /*********************************************************/
  //initialisation de x, X, Y, W et Pn par la mthode recoit x.
  if( x->base )    ambipan_tilde_recoit_x( x, Xdefaut);
  else             ambipan_tilde_recoit_x( x, Rdefaut);     




  /*********************************************************/
  /*Cration des nouvelles entres. ************************/
  //si controle on cre des entres de contrle,
  if( x->controle ) {
    dsp_setup( (t_pxobject *)x, 1);
    floatin(x,1);
    floatin(x,2);
  }	
  //Sinon on cre des entres de signal.
  else   dsp_setup( (t_pxobject *)x, 3);
    
  /*Cration des sorties************************************/
  for( i=0; i < x->N; i++)
    outlet_new((t_object *)x, "signal");

  return (void *)x;
}



/**********************************************************************/
/*                    DEFINITION DE LA CLASSE	                      */
/**********************************************************************/

void main(void) 
{
  /****Dfinition de la classe**********************************/
  setup(	(t_messlist **) &ambipan_tilde_class, 
			    (method) ambipan_tilde_new, 
			    (method) ambipan_tilde_dest, 	
			    (short)sizeof(t_ambipan_tilde), 	
			    0L, 
			    A_GIMME,
			    0);

  /****Dfinition des mthodes**********************************/
  addmess( (method)ambipan_tilde_dsp, "dsp", A_CANT, 0);
  dsp_initclass(  );
  

  /*Mthodes de la classe***************************************/
  addmess( (method) ambipan_tilde_recoit_liste, "list", A_GIMME, 0);
  addfloat((method) ambipan_tilde_recoit_x);
  addftx(  (method) ambipan_tilde_recoit_x , 2);
  addftx(  (method) ambipan_tilde_recoit_y , 1);
  addmess( (method) ambipan_tilde_initialiser_nb_hp, "set_nb_hp", A_LONG, 0);
  addmess( (method) ambipan_tilde_teta_positionner_hp, "a_setpos", A_GIMME, 0);
  addmess( (method) ambipan_tilde_dist_teta_positionner_hp, "ra_setpos", A_GIMME, 0);
  addmess( (method) ambipan_tilde_xy_positionner_hp, "xy_setpos", A_GIMME, 0);
  addmess( (method) ambipan_tilde_changer_offset, "set_offset", A_GIMME, 0);
  addmess( (method) ambipan_tilde_muter_entrees_signal, "mute_sig", A_LONG, 0);
  addmess( (method) ambipan_tilde_changer_type_repere, "change_type", A_SYM, 0);
  addmess( (method) ambipan_tilde_informations, "get_info", 0);


  /*Publicit pour le CICM *********************************/
  post("\"AMBIPAN~\"  Auteurs: R.MIGNOT  "
       "et B.COURRIBET, CICM Universit Paris8,");
  post("           MSH Paris Nord, ACI Jeunes "
       "Chercheurs \"Espaces Sonores\".");
  post("           cicm\100univ-paris8.fr.");
  
  /*Ajout du nom de l'objet dans liste*/
  finder_addclass( "CICMTOOLS", "ambicube~");
  finder_addclass( "CICMTOOLS", "ambipan~");
  finder_addclass( "CICMTOOLS", "vbapan~");
  finder_addclass( "CICMTOOLS", "six2head~");
  finder_addclass( "CICMTOOLS", "cube2head~");
}




/**********************************************************************/
/*                       METHODE DSP                                  */
/**********************************************************************/

void ambipan_tilde_dsp(t_ambipan_tilde *x, t_signal **sp, short *count)
{
  if( x->controle )
  {
    switch( x->Nout )
    {
    case 2:
      dsp_add(ambipan_tilde_perform_controle, 5, x,
              sp[0]->s_vec,   //son d'entre,
              sp[1]->s_vec,   //tous les sons de sortie, ...
              sp[2]->s_vec,   // .........
              sp[0]->s_n);    //taille des blocs.
      break;
    case 3:
      dsp_add(ambipan_tilde_perform_controle, 6, x,
              sp[0]->s_vec, sp[1]->s_vec, sp[2]->s_vec, 
              sp[3]->s_vec, sp[0]->s_n); 
      break;      
    case 4:  
      dsp_add(ambipan_tilde_perform_controle, 7, x,
              sp[0]->s_vec, sp[1]->s_vec, sp[2]->s_vec,
              sp[3]->s_vec, sp[4]->s_vec, sp[0]->s_n); 
      break;      
    case 5:
      dsp_add(ambipan_tilde_perform_controle, 8, x,
              sp[0]->s_vec, sp[1]->s_vec, sp[2]->s_vec,
              sp[3]->s_vec, sp[4]->s_vec, sp[5]->s_vec,
              sp[0]->s_n); 
      break;      
    case 6:
      dsp_add(ambipan_tilde_perform_controle, 9, x,
              sp[0]->s_vec, sp[1]->s_vec, sp[2]->s_vec,
              sp[3]->s_vec, sp[4]->s_vec, sp[5]->s_vec,
              sp[6]->s_vec, sp[0]->s_n); 
      break;      
    case 7:
      dsp_add(ambipan_tilde_perform_controle, 10, x,
              sp[0]->s_vec, sp[1]->s_vec, sp[2]->s_vec,
              sp[3]->s_vec, sp[4]->s_vec, sp[5]->s_vec,
              sp[6]->s_vec, sp[7]->s_vec, sp[0]->s_n); 
      break;
    case 8:
      dsp_add(ambipan_tilde_perform_controle, 11, x,
              sp[0]->s_vec, sp[1]->s_vec, sp[2]->s_vec,
              sp[3]->s_vec, sp[4]->s_vec, sp[5]->s_vec,
              sp[6]->s_vec, sp[7]->s_vec, sp[8]->s_vec,
              sp[0]->s_n); 
      break;
    case 9:
      dsp_add(ambipan_tilde_perform_controle, 12, x,
              sp[0]->s_vec, sp[1]->s_vec, sp[2]->s_vec,
              sp[3]->s_vec, sp[4]->s_vec, sp[5]->s_vec,
              sp[6]->s_vec, sp[7]->s_vec, sp[8]->s_vec,
              sp[9]->s_vec, sp[0]->s_n); 
      break;
    case 10:
      dsp_add(ambipan_tilde_perform_controle, 13, x,
              sp[0]->s_vec, sp[1]->s_vec, sp[2]->s_vec,
              sp[3]->s_vec, sp[4]->s_vec, sp[5]->s_vec,
              sp[6]->s_vec, sp[7]->s_vec, sp[8]->s_vec,
              sp[9]->s_vec, sp[10]->s_vec, sp[0]->s_n); 
      break;
    case 11:
      dsp_add(ambipan_tilde_perform_controle, 14, x,
              sp[0]->s_vec, sp[1]->s_vec, sp[2]->s_vec,
              sp[3]->s_vec, sp[4]->s_vec, sp[5]->s_vec,
              sp[6]->s_vec, sp[7]->s_vec, sp[8]->s_vec,
              sp[9]->s_vec, sp[10]->s_vec, sp[11]->s_vec, 
              sp[0]->s_n); 
      break;
    case 12:
      dsp_add(ambipan_tilde_perform_controle, 15, x,
              sp[0]->s_vec, sp[1]->s_vec, sp[2]->s_vec,
              sp[3]->s_vec, sp[4]->s_vec, sp[5]->s_vec,
              sp[6]->s_vec, sp[7]->s_vec, sp[8]->s_vec,
              sp[9]->s_vec, sp[10]->s_vec, sp[11]->s_vec, 
              sp[12]->s_vec, sp[0]->s_n); 
      break;
    case 13:
      dsp_add(ambipan_tilde_perform_controle, 16, x,
              sp[0]->s_vec, sp[1]->s_vec, sp[2]->s_vec,
              sp[3]->s_vec, sp[4]->s_vec, sp[5]->s_vec,
              sp[6]->s_vec, sp[7]->s_vec, sp[8]->s_vec,
              sp[9]->s_vec, sp[10]->s_vec, sp[11]->s_vec, 
              sp[12]->s_vec, sp[13]->s_vec, sp[0]->s_n); 
      break;
    case 14:
      dsp_add(ambipan_tilde_perform_controle, 17, x,
              sp[0]->s_vec, sp[1]->s_vec, sp[2]->s_vec,
              sp[3]->s_vec, sp[4]->s_vec, sp[5]->s_vec,
              sp[6]->s_vec, sp[7]->s_vec, sp[8]->s_vec,
              sp[9]->s_vec, sp[10]->s_vec, sp[11]->s_vec, 
              sp[12]->s_vec, sp[13]->s_vec, sp[14]->s_vec, 
              sp[0]->s_n); 
      break;
    case 15:
      dsp_add(ambipan_tilde_perform_controle, 18, x,
              sp[0]->s_vec, sp[1]->s_vec, sp[2]->s_vec,
              sp[3]->s_vec, sp[4]->s_vec, sp[5]->s_vec,
              sp[6]->s_vec, sp[7]->s_vec, sp[8]->s_vec,
              sp[9]->s_vec, sp[10]->s_vec, sp[11]->s_vec, 
              sp[12]->s_vec, sp[13]->s_vec, sp[14]->s_vec, 
              sp[15]->s_vec, sp[0]->s_n); 
      break;
    case 16:
      dsp_add(ambipan_tilde_perform_controle, 19, x,
              sp[0]->s_vec, sp[1]->s_vec, sp[2]->s_vec,
              sp[3]->s_vec, sp[4]->s_vec, sp[5]->s_vec,
              sp[6]->s_vec, sp[7]->s_vec, sp[8]->s_vec,
              sp[9]->s_vec, sp[10]->s_vec, sp[11]->s_vec, 
              sp[12]->s_vec, sp[13]->s_vec, sp[14]->s_vec, 
              sp[15]->s_vec, sp[16]->s_vec, sp[0]->s_n); 
      break;
    }
  }
  else
  {
    switch( x->Nout )
    {
    case 2:
      dsp_add(ambipan_tilde_perform_signal, 7, x,
              sp[0]->s_vec,   //son d'entre,
              sp[1]->s_vec,   //entre de x,
              sp[2]->s_vec,   //entre de y,
              sp[3]->s_vec,   //tous les sons de sortie, ...
              sp[4]->s_vec,   // .........
              sp[0]->s_n);    //taille des blocs.
      break;
    case 3:
      dsp_add(ambipan_tilde_perform_signal, 8, x,
              sp[0]->s_vec, sp[1]->s_vec, sp[2]->s_vec,
              sp[3]->s_vec, sp[4]->s_vec, sp[5]->s_vec, 
              sp[0]->s_n);  
      break;      
    case 4:
      dsp_add(ambipan_tilde_perform_signal, 9, x,
              sp[0]->s_vec, sp[1]->s_vec, sp[2]->s_vec,
              sp[3]->s_vec, sp[4]->s_vec, sp[5]->s_vec, 
              sp[6]->s_vec, sp[0]->s_n);  
      break;      
    case 5:
      dsp_add(ambipan_tilde_perform_signal, 10, x,
              sp[0]->s_vec, sp[1]->s_vec, sp[2]->s_vec,
              sp[3]->s_vec, sp[4]->s_vec, sp[5]->s_vec, 
              sp[6]->s_vec, sp[7]->s_vec, sp[0]->s_n);  
      break;      
    case 6:
      dsp_add(ambipan_tilde_perform_signal, 11, x,
              sp[0]->s_vec, sp[1]->s_vec, sp[2]->s_vec,
              sp[3]->s_vec, sp[4]->s_vec, sp[5]->s_vec, 
              sp[6]->s_vec, sp[7]->s_vec, sp[8]->s_vec, 
              sp[0]->s_n);
      break;      
    case 7:
      dsp_add(ambipan_tilde_perform_signal, 12, x,
              sp[0]->s_vec, sp[1]->s_vec, sp[2]->s_vec,
              sp[3]->s_vec, sp[4]->s_vec, sp[5]->s_vec, 
              sp[6]->s_vec, sp[7]->s_vec, sp[8]->s_vec, 
              sp[9]->s_vec, sp[0]->s_n);  
      break;      
    case 8:
      dsp_add(ambipan_tilde_perform_signal, 13, x,
              sp[0]->s_vec, sp[1]->s_vec, sp[2]->s_vec,
              sp[3]->s_vec, sp[4]->s_vec, sp[5]->s_vec, 
              sp[6]->s_vec, sp[7]->s_vec, sp[8]->s_vec, 
              sp[9]->s_vec, sp[10]->s_vec, sp[0]->s_n);  
      break;      
    case 9:
      dsp_add(ambipan_tilde_perform_signal, 14, x,
              sp[0]->s_vec, sp[1]->s_vec, sp[2]->s_vec,
              sp[3]->s_vec, sp[4]->s_vec, sp[5]->s_vec, 
              sp[6]->s_vec, sp[7]->s_vec, sp[8]->s_vec, 
              sp[9]->s_vec, sp[10]->s_vec, sp[11]->s_vec, 
              sp[0]->s_n);  
      break;
    case 10:
      dsp_add(ambipan_tilde_perform_signal, 15, x,
              sp[0]->s_vec, sp[1]->s_vec, sp[2]->s_vec,
              sp[3]->s_vec, sp[4]->s_vec, sp[5]->s_vec, 
              sp[6]->s_vec, sp[7]->s_vec, sp[8]->s_vec, 
              sp[9]->s_vec, sp[10]->s_vec, sp[11]->s_vec, 
              sp[11]->s_vec, sp[0]->s_n);  
      break;
    case 11:
      dsp_add(ambipan_tilde_perform_signal, 16, x,
              sp[0]->s_vec, sp[1]->s_vec, sp[2]->s_vec,
              sp[3]->s_vec, sp[4]->s_vec, sp[5]->s_vec, 
              sp[6]->s_vec, sp[7]->s_vec, sp[8]->s_vec, 
              sp[9]->s_vec, sp[10]->s_vec, sp[11]->s_vec, 
              sp[12]->s_vec, sp[13]->s_vec, sp[0]->s_n);  
      break;
    case 12:
      dsp_add(ambipan_tilde_perform_signal, 17, x,
              sp[0]->s_vec, sp[1]->s_vec, sp[2]->s_vec,
              sp[3]->s_vec, sp[4]->s_vec, sp[5]->s_vec, 
              sp[6]->s_vec, sp[7]->s_vec, sp[8]->s_vec, 
              sp[9]->s_vec, sp[10]->s_vec, sp[11]->s_vec, 
              sp[12]->s_vec, sp[13]->s_vec, sp[14]->s_vec, 
              sp[0]->s_n);  
      break;
    case 13:
      dsp_add(ambipan_tilde_perform_signal, 18, x,
              sp[0]->s_vec, sp[1]->s_vec, sp[2]->s_vec,
              sp[3]->s_vec, sp[4]->s_vec, sp[5]->s_vec, 
              sp[6]->s_vec, sp[7]->s_vec, sp[8]->s_vec, 
              sp[9]->s_vec, sp[10]->s_vec, sp[11]->s_vec, 
              sp[12]->s_vec, sp[13]->s_vec, sp[14]->s_vec, 
              sp[15]->s_vec, sp[0]->s_n);  
      break;
    case 14:
      dsp_add(ambipan_tilde_perform_signal, 19, x,
              sp[0]->s_vec, sp[1]->s_vec, sp[2]->s_vec,
              sp[3]->s_vec, sp[4]->s_vec, sp[5]->s_vec, 
              sp[6]->s_vec, sp[7]->s_vec, sp[8]->s_vec, 
              sp[9]->s_vec, sp[10]->s_vec, sp[11]->s_vec, 
              sp[12]->s_vec, sp[13]->s_vec, sp[14]->s_vec, 
              sp[15]->s_vec, sp[16]->s_vec, sp[0]->s_n);  
      break;
    case 15:
      dsp_add(ambipan_tilde_perform_signal, 20, x,
              sp[0]->s_vec, sp[1]->s_vec, sp[2]->s_vec,
              sp[3]->s_vec, sp[4]->s_vec, sp[5]->s_vec, 
              sp[6]->s_vec, sp[7]->s_vec, sp[8]->s_vec, 
              sp[9]->s_vec, sp[10]->s_vec, sp[11]->s_vec, 
              sp[12]->s_vec, sp[13]->s_vec, sp[14]->s_vec, 
              sp[15]->s_vec, sp[16]->s_vec, sp[17]->s_vec, 
              sp[0]->s_n);  
      break;
    case 16:
      dsp_add(ambipan_tilde_perform_signal, 21, x,
              sp[0]->s_vec, sp[1]->s_vec, sp[2]->s_vec,
              sp[3]->s_vec, sp[4]->s_vec, sp[5]->s_vec, 
              sp[6]->s_vec, sp[7]->s_vec, sp[8]->s_vec, 
              sp[9]->s_vec, sp[10]->s_vec, sp[11]->s_vec, 
              sp[12]->s_vec, sp[13]->s_vec, sp[14]->s_vec, 
              sp[15]->s_vec, sp[16]->s_vec, sp[17]->s_vec, 
              sp[18]->s_vec, sp[0]->s_n);  
      break;
    }
  }
}